/*
 * Win32Service.java
 *
 * Created on 12. September 2007, 12:05
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package jnacontrib.win32;

import java.util.HashMap;
import java.util.Map;

import jnacontrib.jna.Advapi32;
import jnacontrib.jna.WINERROR;
import jnacontrib.jna.WINNT;
import jnacontrib.jna.WINSVC;
import jnacontrib.jna.Advapi32.ENUM_SERVICE_STATUS_PROCESS;
import jnacontrib.jna.Advapi32.QUERY_SERVICE_CONFIG;
import jnacontrib.jna.Advapi32.SERVICE_STATUS_PROCESS;



import org.rzo.yajsw.os.Service;
import org.rzo.yajsw.os.ms.win.w32.WindowsXPProcess.MyKernel32;

import com.sun.jna.Memory;
import com.sun.jna.Pointer;
import com.sun.jna.examples.win32.Kernel32;
import com.sun.jna.ptr.IntByReference;
import com.sun.jna.ptr.PointerByReference;

/**
 * Baseclass for a Win32 service.
 */
abstract public class Win32Service
{
	protected String		serviceName;
	private ServiceMain		serviceMain;
	private ServiceControl	serviceControl;
	private Pointer			serviceStatusHandle;
	private Object			waitObject	= new Object();

	/**
	 * Creates a new instance of Win32Service.
	 * 
	 * @param serviceName
	 *            internal name of the service
	 */
	public Win32Service(String serviceName)
	{
		this.serviceName = serviceName;
	}

	public Win32Service()
	{

	}

	public void setServiceName(String serviceName)
	{
		this.serviceName = serviceName;
	}

	public String getServiceName()
	{
		return serviceName;
	}

	/**
	 * Install the service.
	 * 
	 * @param displayName
	 *            visible name
	 * @param description
	 *            description
	 * @param dependencies
	 *            array of other services to depend on or null
	 * @param account
	 *            service account or null for LocalSystem
	 * @param password
	 *            password for service account or null
	 * @throws java.lang.Exception
	 * @return true on success
	 */
	public boolean install(String displayName, String description, String[] dependencies, String account, String password)
	{
		return (install(displayName, description, dependencies, account, password, "java.exe -cp \"" + System.getProperty("java.class.path")
				+ "\" -Xrs " + this.getClass().getName(), "AUTO_START", false));
	}

	/**
	 * Install the service.
	 * 
	 * @return true on success
	 * @param displayName
	 *            visible name
	 * @param description
	 *            description
	 * @param dependencies
	 *            array of other services to depend on or null
	 * @param account
	 *            service account or null for LocalSystem
	 * @param password
	 *            password for service account or null
	 * @param command
	 *            command line to start the service
	 * @throws java.lang.Exception
	 */
	public boolean install(String displayName, String description, String[] dependencies, String account, String password, String command,
			String startType, boolean interactive)
	{
		Advapi32 advapi32;
		Advapi32.SERVICE_DESCRIPTION desc;
		Pointer serviceManager, service;
		boolean success = false;
		String dep = "";

		if (dependencies != null)
		{
			for (String s : dependencies)
			{
				dep += s + "\0";
			}
		}
		dep += "\0";

		desc = new Advapi32.SERVICE_DESCRIPTION();
		desc.lpDescription = description;

		advapi32 = Advapi32.INSTANCE;
		serviceManager = openServiceControlManager(null, WINSVC.SC_MANAGER_ALL_ACCESS);

		int winStartType = "DEMAND_START".equals(startType) ? WINSVC.SERVICE_DEMAND_START : WINSVC.SERVICE_AUTO_START;
		int dwServiceType = WINSVC.SERVICE_WIN32_OWN_PROCESS;
		if (interactive)
			dwServiceType |= WINSVC.SERVICE_INTERACTIVE_PROCESS;
		if (serviceManager != null)
		{
			service = advapi32.CreateService(serviceManager, serviceName, displayName, WINSVC.SERVICE_ALL_ACCESS, dwServiceType, winStartType,
					WINSVC.SERVICE_ERROR_NORMAL, command, null, null, dep, account, password);

			if (service != null)
			{
				success = advapi32.ChangeServiceConfig2(service, WINSVC.SERVICE_CONFIG_DESCRIPTION, desc);
				advapi32.CloseServiceHandle(service);
			}
			else
			{
				int err = Kernel32.INSTANCE.GetLastError();
				System.out.println("error during install "+err);
				System.out.println(getSystemError(err));
			}
				
			advapi32.CloseServiceHandle(serviceManager);
		}
		return (success);
	}
	
    private String getSystemError(int code) {
        Kernel32 lib = Kernel32.INSTANCE;
        PointerByReference pref = new PointerByReference();
        lib.FormatMessage(Kernel32.FORMAT_MESSAGE_ALLOCATE_BUFFER
                          |Kernel32.FORMAT_MESSAGE_FROM_SYSTEM
                          |Kernel32.FORMAT_MESSAGE_IGNORE_INSERTS, 
                          null, code, 
                          0, pref, 0, null);
        String s = pref.getValue().getString(0, !Boolean.getBoolean("w32.ascii"));
        s = s.replace(".\r",".").replace(".\n",".");
        lib.LocalFree(pref.getValue());
        return s;
    }


	/**
	 * Uninstall the service.
	 * 
	 * @throws java.lang.Exception
	 * @return true on success
	 */
	public boolean uninstall()
	{
		Advapi32 advapi32;
		Pointer serviceManager, service;
		boolean success = false;

		advapi32 = Advapi32.INSTANCE;
		serviceManager = openServiceControlManager(null, WINSVC.SC_MANAGER_ALL_ACCESS);

		if (serviceManager != null)
		{
			service = advapi32.OpenService(serviceManager, serviceName, WINSVC.SERVICE_ALL_ACCESS);

			if (service != null)
			{
				success = advapi32.DeleteService(service);
				advapi32.CloseServiceHandle(service);
			}
			advapi32.CloseServiceHandle(serviceManager);
		}
		return (success);
	}

	public int state()
	{
		Advapi32 advapi32;
		Pointer serviceManager, service;
		int result = Service.STATE_UNKNOWN;

		advapi32 = Advapi32.INSTANCE;

		serviceManager = openServiceControlManager(null, WINNT.GENERIC_READ);

		if (serviceManager != null)
		{
			result = 0;
			service = advapi32.OpenService(serviceManager, serviceName, WINNT.GENERIC_READ);

			if (service != null)
			{
				IntByReference pcbBytesNeeded = new IntByReference();
				result |= Service.STATE_INSTALLED;
				// get size required
				if (!advapi32.QueryServiceConfig(service, null, 0, pcbBytesNeeded))
				{
					// now get the data
					int cbBufSize = pcbBytesNeeded.getValue();
					Memory buffer = new Memory(cbBufSize);
					buffer.clear();
					if (advapi32.QueryServiceConfig(service, buffer, cbBufSize, pcbBytesNeeded))
					{
						QUERY_SERVICE_CONFIG lpServiceConfig = new QUERY_SERVICE_CONFIG();
						lpServiceConfig.init(buffer);
						if (lpServiceConfig.dwStartType == Advapi32.SERVICE_DISABLED)
							result |= Service.STATE_DISABLED;
						if (lpServiceConfig.dwStartType == Advapi32.SERVICE_BOOT_START | lpServiceConfig.dwStartType == Advapi32.SERVICE_SYSTEM_START
								| lpServiceConfig.dwStartType == Advapi32.SERVICE_AUTO_START)
							result |= Service.STATE_AUTOMATIC;
						if (lpServiceConfig.dwStartType == Advapi32.SERVICE_DEMAND_START)
							result |= Service.STATE_MANUAL;
						if ((lpServiceConfig.dwServiceType & Advapi32.SERVICE_INTERACTIVE_PROCESS) != 0)
							result |= Service.STATE_INTERACTIVE;

					}
					else
					{
						result |= Service.STATE_UNKNOWN;
						System.out.println("Error in QueryServiceConfig: " + MyKernel32.INSTANCE.GetLastError());
					}
				}
				else
				{
					result |= Service.STATE_UNKNOWN;
					System.out.println("Error in QueryServiceConfig: " + MyKernel32.INSTANCE.GetLastError());
				}
				if (!advapi32.QueryServiceStatusEx(service, (byte) advapi32.SC_STATUS_PROCESS_INFO, null, 0, pcbBytesNeeded))
				{
					// now get the data
					int cbBufSize = pcbBytesNeeded.getValue();
					Memory buffer = new Memory(cbBufSize);
					buffer.clear();
					if (advapi32.QueryServiceStatusEx(service, (byte) advapi32.SC_STATUS_PROCESS_INFO, buffer, cbBufSize, pcbBytesNeeded))
					{
						SERVICE_STATUS_PROCESS lpBuffer = new SERVICE_STATUS_PROCESS();
						lpBuffer.init(buffer);
						if (lpBuffer.dwCurrentState == advapi32.SERVICE_RUNNING)
							result |= Service.STATE_RUNNING;
						if (lpBuffer.dwCurrentState == advapi32.SERVICE_PAUSED)
							result |= Service.STATE_PAUSED;
					}
					else
					{
						result |= Service.STATE_UNKNOWN;
						System.out.println("Error in QueryServiceStatusEx: " + MyKernel32.INSTANCE.GetLastError());
					}
				}
				else
				{
					result |= Service.STATE_UNKNOWN;
					System.out.println("Error in QueryServiceStatusEx: " + MyKernel32.INSTANCE.GetLastError());
				}

				advapi32.CloseServiceHandle(service);
			}
			advapi32.CloseServiceHandle(serviceManager);
		}

		return result;

	}

	/**
	 * Ask the ServiceControlManager to start the service.
	 * 
	 * @return true on success
	 */
	public boolean start()
	{
		Advapi32 advapi32;
		Pointer serviceManager, service;
		boolean success = false;

		advapi32 = Advapi32.INSTANCE;

		serviceManager = openServiceControlManager(null, WINNT.GENERIC_EXECUTE);

		if (serviceManager != null)
		{
			service = advapi32.OpenService(serviceManager, serviceName, WINNT.GENERIC_EXECUTE);

			if (service != null)
			{
				success = advapi32.StartService(service, 0, null);
				advapi32.CloseServiceHandle(service);
			}
			advapi32.CloseServiceHandle(serviceManager);
		}

		return (success);
	}

	/**
	 * Ask the ServiceControlManager to stop the service.
	 * 
	 * @return true on success
	 */
	public boolean stop() throws Exception
	{
		Advapi32 advapi32;
		Pointer serviceManager, service;
		Advapi32.SERVICE_STATUS serviceStatus;
		boolean success = false;

		advapi32 = Advapi32.INSTANCE;

		serviceManager = openServiceControlManager(null, WINNT.GENERIC_EXECUTE);

		if (serviceManager != null)
		{
			service = advapi32.OpenService(serviceManager, serviceName, WINNT.GENERIC_EXECUTE);

			if (service != null)
			{
				serviceStatus = new Advapi32.SERVICE_STATUS();
				success = advapi32.ControlService(service, WINSVC.SERVICE_CONTROL_STOP, serviceStatus);
				advapi32.CloseServiceHandle(service);
			}
			advapi32.CloseServiceHandle(serviceManager);
		}

		return (success);
	}

	/**
	 * Initialize the service, connect to the ServiceControlManager.
	 */
	public void init()
	{
		Advapi32 advapi32;
		Advapi32.SERVICE_TABLE_ENTRY[] entries = new Advapi32.SERVICE_TABLE_ENTRY[2];
		Advapi32.SERVICE_TABLE_ENTRY entry;

		serviceMain = new ServiceMain();
		advapi32 = Advapi32.INSTANCE;
		entry = new Advapi32.SERVICE_TABLE_ENTRY();
		entry.lpServiceName = serviceName;
		entry.lpServiceProc = serviceMain;

		advapi32.StartServiceCtrlDispatcher(entry.toArray(2));
	}

	/**
	 * Get a handle to the ServiceControlManager.
	 * 
	 * @param machine
	 *            name of the machine or null for localhost
	 * @param access
	 *            access flags
	 * @return handle to ServiceControlManager or null when failed
	 */
	static private Pointer openServiceControlManager(String machine, int access)
	{
		Pointer handle = null;
		Advapi32 advapi32;

		advapi32 = Advapi32.INSTANCE;
		handle = advapi32.OpenSCManager(machine, null, access);
		if (handle == null)
		{
			int err = MyKernel32.INSTANCE.GetLastError();
			System.out.println("Error in OpenSCManager: " + Integer.toHexString(err));
			if (err == 5)
				System.out.println("Access denied: please check the user credentials");
		}

		return (handle);
	}

	static public Map<String, ENUM_SERVICE_STATUS_PROCESS> enumerateServices(String machine)
	{
		Map<String, ENUM_SERVICE_STATUS_PROCESS> result = new HashMap();
		// Open the Service Control Manager
		Pointer sc = openServiceControlManager(machine, WINSVC.SC_MANAGER_ENUMERATE_SERVICE);

		// Check if OpenSCManager returns NULL. Otherwise proceed
		if (sc != null && !sc.equals(null))
		{
			Memory service_data = null;
			int service_data_size = 0;
			int infoLevel = WINSVC.SC_ENUM_PROCESS_INFO;
			boolean retVal;
			IntByReference bytesNeeded = new IntByReference(0);
			IntByReference srvCount = new IntByReference(0);
			IntByReference resumeHandle = new IntByReference(0);
			int srvType = WINSVC.SERVICE_WIN32;
			int srvState = WINSVC.SERVICE_STATE_ALL;

			// Call EnumServicesStatus with null data and data_size == 0, so we
			// get the required memory size
			retVal = Advapi32.INSTANCE.EnumServicesStatusExW(sc, infoLevel, srvType, srvState, service_data, service_data_size, bytesNeeded,
					srvCount, resumeHandle, null);

			int err = MyKernel32.INSTANCE.GetLastError();
			// EnumServicesStatus should need more memory space
			if ((!retVal) || err == WINERROR.ERROR_MORE_DATA)
			{
				int bytesCount = bytesNeeded.getValue();
				service_data = new Memory(bytesCount);
				service_data.clear();
				service_data_size = bytesCount;
				System.out.println(resumeHandle.getValue());
				resumeHandle.setValue(0);
				retVal = Advapi32.INSTANCE.EnumServicesStatusExW(sc, infoLevel, srvType, srvState, service_data, service_data_size, bytesNeeded,
						srvCount, resumeHandle, null);
				if (!retVal)
				{
					err = MyKernel32.INSTANCE.GetLastError();
					System.out.println("Error in EnumServicesStatusExA " + Integer.toHexString(err));
					return null;
				}
			}
			else
				return null;

			ENUM_SERVICE_STATUS_PROCESS serviceStatus = new ENUM_SERVICE_STATUS_PROCESS();
			serviceStatus.init(service_data);
			for (int i = 0; i < srvCount.getValue(); i++)
			{
				result.put(serviceStatus.getServiceName().toLowerCase(), serviceStatus);
				serviceStatus = serviceStatus.next();
			}
		}

		// Close the SC_HANLDE returned by OpenSCManager
		Advapi32.INSTANCE.CloseServiceHandle(sc);

		return result;
	}

	/**
	 * Report service status to the ServiceControlManager.
	 * 
	 * @param status
	 *            status
	 * @param win32ExitCode
	 *            exit code
	 * @param waitHint
	 *            time to wait
	 */
	private void reportStatus(int status, int win32ExitCode, int waitHint)
	{
		Advapi32 advapi32;
		Advapi32.SERVICE_STATUS serviceStatus;

		advapi32 = Advapi32.INSTANCE;
		serviceStatus = new Advapi32.SERVICE_STATUS();
		serviceStatus.dwServiceType = WINNT.SERVICE_WIN32_OWN_PROCESS;
		serviceStatus.dwControlsAccepted = WINSVC.SERVICE_ACCEPT_STOP | WINSVC.SERVICE_ACCEPT_SHUTDOWN;
		serviceStatus.dwWin32ExitCode = win32ExitCode;
		serviceStatus.dwWaitHint = waitHint;
		serviceStatus.dwCurrentState = status;

		advapi32.SetServiceStatus(serviceStatusHandle, serviceStatus);
	}

	/**
	 * Called when service is starting.
	 */
	public abstract void onStart();

	/*
	 * Called when service should stop.
	 */
	public abstract void onStop();

	/**
	 * Implementation of the service main function.
	 */
	private class ServiceMain implements Advapi32.SERVICE_MAIN_FUNCTION
	{

		/**
		 * Called when the service is starting.
		 * 
		 * @param dwArgc
		 *            number of arguments
		 * @param lpszArgv
		 *            pointer to arguments
		 */
		public void callback(int dwArgc, Pointer lpszArgv)
		{
			Advapi32 advapi32;

			advapi32 = Advapi32.INSTANCE;

			serviceControl = new ServiceControl();
			serviceStatusHandle = advapi32.RegisterServiceCtrlHandlerEx(serviceName, serviceControl, null);

			reportStatus(WINSVC.SERVICE_START_PENDING, WINERROR.NO_ERROR, 3000);
			reportStatus(WINSVC.SERVICE_RUNNING, WINERROR.NO_ERROR, 0);

			onStart();

			try
			{
				synchronized (waitObject)
				{
					waitObject.wait();
				}
			}
			catch (InterruptedException ex)
			{
			}
			reportStatus(WINSVC.SERVICE_STOPPED, WINERROR.NO_ERROR, 0);

			// Avoid returning from ServiceMain, which will cause a crash
			// See http://support.microsoft.com/kb/201349, which recommends
			// having init() wait for this thread.
			// Waiting on this thread in init() won't fix the crash, though.
			// System.exit(0);
		}
	}

	/**
	 * Implementation of the service control function.
	 */
	private class ServiceControl implements Advapi32.HandlerEx
	{

		/**
		 * Called when the service get a control code.
		 * 
		 * @param dwControl
		 * @param dwEventType
		 * @param lpEventData
		 * @param lpContext
		 */
		public int callback(int dwControl, int dwEventType, Pointer lpEventData, Pointer lpContext)
		{
			switch (dwControl)
			{
			case WINSVC.SERVICE_CONTROL_STOP:
			case WINSVC.SERVICE_CONTROL_SHUTDOWN:
				reportStatus(WINSVC.SERVICE_STOP_PENDING, WINERROR.NO_ERROR, 5000);
				onStop();
				synchronized (waitObject)
				{
					waitObject.notifyAll();
				}
			}
			return WINERROR.NO_ERROR;
		}
	}
}
